﻿// This source code is property of the Computer Graphics and Visualization
// chair of the TU Dresden. Do not distribute!
// Copyright (C) CGV TU Dresden - All Rights Reserved
//
#include "Bone.h"

#include <cgv/math/transformations.h>
#include <cgv/math/inv.h>

#include "math_helper.h"

float calc_angle(Vec3 dir_vec, Vec3 axis)
{
	float dot_result = dot(dir_vec, axis);

	float length_v1 = length(dir_vec);
	float length_v2 = length(axis);

	return acos(dot_result / (length_v1 * length_v2));
}

Bone::Bone()
	: parent(nullptr), length(0.0f), direction_in_world_space(0.0, 0.0, 0.0), translationTransforms(0)
{
}

Bone::~Bone()
{
	for (auto bone : children)
		delete bone;
	children.clear();

	dofs.clear();

	for (auto dof : orientation)
		delete dof;
	orientation.clear();
}

void Bone::calculate_matrices()
{
	orientationTransformGlobalToLocal.identity();
	std::for_each(orientation.begin(), orientation.end(), [&](AtomicTransform *t) {
		std::cout << "atomic transform " << t->get_name() << ": " << t->get_value() << std::endl;

		orientationTransformGlobalToLocal = orientationTransformGlobalToLocal * t->calculate_matrix();
	});
	orientationTransformLocalToGlobal = cgv::math::inv(orientationTransformGlobalToLocal);

	/*Task 2.1: Implement matrix calculation */
	Mat4 matrix;
	matrix.identity();

	Vec3 vector = direction_in_world_space * length;

	translationTransformCurrentJointToNext = translate(vector);
	orientationTransformPrevJointToCurrent = matrix;

	/*
	translationTransformCurrentJointToNext = translate(0.0f, 0.0f, length);

	if (parent == NULL)
	{
		if (direction_in_world_space.x() == 0.0f &&
		direction_in_world_space.y() == 0.0f &&
		direction_in_world_space.z() == 0.0f) {
			orientationTransformPrevJointToCurrent = rotate(Vec3(0.0f, 0.0f, 1.0f), 0.0f);
		} else {
std::cout << direction_in_world_space.x() << ", "
				<< direction_in_world_space.y() << ", "
				<< direction_in_world_space.z() << std::endl;

		float z_angle = calc_angle(direction_in_world_space, Vec3(0.0f, 0.0f, 1.0f));
		float y_angle = calc_angle(direction_in_world_space, Vec3(0.0f, 1.0f, 0.0f));
		float x_angle = calc_angle(direction_in_world_space, Vec3(1.0f, 0.0f, 0.0f));

		std::cout << "z_angle " << z_angle << std::endl;
		std::cout << "y_angle " << y_angle << std::endl;
		std::cout << "x_angle " << x_angle << std::endl;

			Mat4 z_mat = rotate(Vec3(0.0f, 0.0f, 1.0f), z_angle);
			Mat4 y_mat = rotate(Vec3(0.0f, 1.0f, 0.0f), y_angle);
			Mat4 x_mat = rotate(Vec3(1.0f, 0.0f, 0.0f), x_angle);

			Mat4 result = z_mat * y_mat * x_mat;

			orientationTransformPrevJointToCurrent = result;
		}
	}
	else
	{
		Mat4 parent_orientation = parent->get_orientation_transform_prev_joint_to_current();

		float angle = calc_angle(direction_in_world_space, parent->get_direction_in_world_space());

		Mat4 rot = rotate(Vec3(1.0f, 1.0f, 1.0f), angle);

		orientationTransformPrevJointToCurrent = rot;
	}
	*/

	/*Task 4.6: Implement matrix calculation */
}

Mat4 Bone::calculate_transform_prev_to_current_with_dofs()
{
	//Task 2.1: Implement matrix calculation
	Mat4 dofs;
	dofs.identity();

	std::for_each(orientation.begin(), orientation.end(), [&](AtomicTransform *t) {
		dofs = dofs * t->calculate_matrix();
	});

	Mat4 t = translationTransformCurrentJointToNext * orientationTransformPrevJointToCurrent * dofs;
	return t;
}

Mat4 Bone::calculate_transform_prev_to_current_without_dofs()
{
	//Task 2.1: Implement matrix calculation
	Mat4 t = translationTransformCurrentJointToNext * orientationTransformPrevJointToCurrent;
	return t;
}

void Bone::add_dof(AtomicTransform *dof)
{
	dof->set_index_in_amc(dofs.size());
	if (dynamic_cast<AtomicTranslationTransform *>(dof))
	{
		dofs.push_front(std::shared_ptr<AtomicTransform>(dof));
		++translationTransforms;
	}
	else
		dofs.insert(dofs.begin() + translationTransforms, std::shared_ptr<AtomicTransform>(dof));
}

void Bone::set_name(const std::string &name) { this->name = name; }
const std::string &Bone::get_name() const { return name; }

void Bone::set_direction_in_world_space(const Vec3 &direction) { this->direction_in_world_space = direction; }
const Vec3 &Bone::get_direction_in_world_space() const { return direction_in_world_space; }

void Bone::set_length(float l) { this->length = l; }
float Bone::get_length() const { return length; }

void Bone::add_axis_rotation(AtomicTransform *transform) { orientation.push_front(transform); }
void Bone::add_child(Bone *child)
{
	child->set_parent(this);
	children.push_back(child);
}
Bone *Bone::child_at(int i) const { return children[i]; }
int Bone::childCount() const { return children.size(); }

void Bone::set_parent(Bone *parent)
{
	this->parent = parent;
}
Bone *Bone::get_parent() const { return parent; }

int Bone::dof_count() const { return dofs.size(); }
std::shared_ptr<AtomicTransform> Bone::get_dof(int dofIndex) const { return dofs[dofIndex]; }

const Mat4 &Bone::get_binding_pose_matrix() const
{
	return transformLocalToGlobal;
}

const Mat4 &Bone::get_translation_transform_current_joint_to_next() const { return translationTransformCurrentJointToNext; }
const Mat4 &Bone::get_orientation_transform_prev_joint_to_current() const { return orientationTransformPrevJointToCurrent; }

Vec4 Bone::get_bone_local_root_position() const { return Vec4(0, 0, 0, 1); }
Vec4 Bone::get_bone_local_tip_position() const { return translationTransformCurrentJointToNext * Vec4(0, 0, 0, 1); }
